/* 
 Author: Jose Romeo Vela        Date: May 4, 2002
 email: jrvela@aristasol.com
 http://www.aristasol.com/

 This software is licensed under GNU's GPL. For details see http://www.gnu.org/

 Copyright (c) 2002, Jose Romeo Vela

==============================================================================

*/



/*
  These utilities are to manipulate the symbol table for the casm assembler.

  The symbol table is implemented as a linked list. To access the elements
  we use the sym_table_search utility.

  To add a defined symbol to the table user sym_table_add.

  Unresolved symbols are handled with the backpatching technique. Symbols that
  don't have a value in the symbol table, keep a linked list to the cardiac 
  that needs the symbol. Once an unresolved symbol becomes resolved, the
  backpatch linked list is used to resolve all the references in the code.
  This technique allows the assembler to be implemented in a single pass.

  If there is a reference to an unresolved symbol that does not exist in the
  symbol table, we use sym_table_unresolved_new to create it.

  If there is a reference to an unresolved symbol that already exists
  in the symbol table, we use sym_table_unresolved_add to add an entry to the
  backpatch linked list.

  If an unresolved symble becomes resolved, we use bkpatch to resolve all
  the references and delete the backpatch linked list for that symbol.

  sym_table_destroy is used to returned the memory once we no longer need the
  symbol table.
  
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "casm.data.h"

FILE * out;

int sym_table_destroy(struct sym_table * s)
{
  struct sym_table * s_del;
  struct backpatch * b;
  struct backpatch * b_del;

  while (s != NULL) {

    b = s->backpatch_ptr ;
    while ( b != NULL ){
      b_del = b;
      b = b->next;
      free(b_del);
    }
    s_del = s;
    s = s->next;
    free(s_del);
  }
}

int sym_table_check(struct sym_table *  s)
{
  int rc = 0;

  while (s != NULL){
    if (s->backpatch_ptr != NULL) { 
      rc++ ; 
      if ( strstr(s->symbol,"#") == NULL ) {
	fprintf(stderr,"Unresolved user symbol: \"%s\"\n",s->symbol);
      }
    };
    s = s->next;
  }

  return(rc);

}

int sym_table_add(struct sym_table * * s,char * symbol,int value)
{
  struct sym_table * p;

  p = malloc(sizeof(struct sym_table));
  p->backpatch_ptr = NULL;
  strcpy(&p->symbol[0],symbol);
  p->value = value;
  p->next = *s;
  *s = p;
}

int sym_table_unresolved_new(struct sym_table * * s,char * symbol,struct cardiac_code * c)
{
    struct sym_table * p;
    struct backpatch * b;

    p = *s;

    sym_table_add(&p, symbol,0);
    
    b= malloc(sizeof(struct backpatch));
    
    p->backpatch_ptr = b;

    b->c_code_ptr = c;
    b->next = NULL;

    *s = p;
    

}

int sym_table_unresolved_add(struct sym_table * s,struct cardiac_code *c)
{
    struct backpatch * b;
    struct backpatch * t;

    b = malloc(sizeof(struct backpatch));
    
    b->next = s->backpatch_ptr;
    b->c_code_ptr = c;
    s->backpatch_ptr = b;

}

int sym_table_search(struct sym_table * * s,char * symbol)
{
  struct sym_table * r;

  r = *s;
  /*  printf("%s\n",r); */
  while ( r != NULL &&  strcmp(r->symbol,symbol)){
    /*    printf("search: %s != %s\n",r,symbol); */
    r=r->next;
  }
  *s = r;
}

int bkpatch(struct sym_table * s)
{
  struct backpatch * b;
  struct backpatch * del;

  b = s->backpatch_ptr;

  while (b != NULL ) {
    b->c_code_ptr->cell = b->c_code_ptr->cell + s->value;
    del = b;
    b = b->next;
    free(del);
  }
  s->backpatch_ptr = NULL;

}

int code_add(struct cardiac_code * * start_ptr, 
	     struct cardiac_code * * end_ptr,
	     int address, 
	     int cell)
{
  struct cardiac_code * c;
  struct cardiac_code * p;

  c = malloc(sizeof(struct cardiac_code));
  c->address=address;
  c->cell = cell;
  c->next = NULL;
  c->unresolved_symbol_ptr=NULL;

  p = *end_ptr;

  if (p != NULL) {
    p->next = c;
  } else {
    *start_ptr = c;
  }
  *end_ptr = c;
}


int code_out(struct cardiac_code * code_ptr)
{
  while (code_ptr != NULL){
    fprintf(out,"%02d:%03d\n",code_ptr->address,code_ptr->cell);
    code_ptr=code_ptr->next;
  }
}

int code_destroy(struct cardiac_code * code_ptr)
{
  struct cardiac_code * t;
  struct cardiac_code * p;
  
  p = code_ptr;

  while (p != NULL){
    t = p;
    p = p->next;
    free(t);
  }

}
